import os
from pathlib import Path
from typing import Dict, Optional

from bgmi.config import cfg
from bgmi.front.base import COVER_URL, BaseHandler
from bgmi.lib.table import STATUS_DELETED, STATUS_END, STATUS_UPDATING, Followed
from bgmi.utils import bangumi_save_path, normalize_path


def get_player(bangumi_name: str) -> Dict[int, Dict[str, str]]:
    bangumi_path = bangumi_save_path(bangumi_name)

    if not bangumi_path.exists():
        return {}

    episode_list = {}

    episodes = [episode.name for episode in bangumi_path.iterdir() if episode.name.isdigit()]

    for episode in episodes:
        e = find_largest_video_file(bangumi_path.joinpath(episode))
        if e:
            episode_list[int(episode)] = {"path": "/" + e}

    return episode_list


def find_largest_video_file(top_dir: Path) -> Optional[str]:
    video_files = []
    for root, _, files in os.walk(top_dir):
        for file in files:
            _, ext = os.path.splitext(file)
            if ext.lower() in [".mp4", ".mkv", ".webm"]:
                p = Path(root).joinpath(file)
                video_files.append((p.stat().st_size, p))

    if not video_files:
        return None

    video_files.sort(key=lambda x: -x[0])

    return video_files[0][1].relative_to(cfg.save_path).as_posix()


if __name__ == "__main__":
    print(get_player("test-save-path"))


class IndexHandler(BaseHandler):
    def get(self, path: str) -> None:
        if not os.path.exists(cfg.front_static_path):
            msg = """<h1>Thanks for your using BGmi</h1>
            <p>It seems you have not install BGmi Frontend,
             please run <code>bgmi install</code> to install.</p>
            """
        else:
            msg = """<h1>Thanks for your using BGmi</h1>
            <p>If use want to use Tornado to serve static files, please enable
            <code>[http]</code>,
            <code>serve_static_files = false</code>,
            and do not forget install bgmi-frontend by
            running <code>bgmi install</code></p>"""

        self.write(msg)
        self.finish()


class BangumiListHandler(BaseHandler):
    def get(self, type_: str = "") -> None:
        data = [
            x if isinstance(x, dict) else x.__dict__
            for x in Followed.get_all_followed(STATUS_DELETED, STATUS_END if type_ == "old" else STATUS_UPDATING)
        ]

        def sorter(_: Dict[str, int]) -> int:
            return _["updated_time"] if _["updated_time"] else 1

        if type_ == "index":
            data.extend(self.patch_list)
            data.sort(key=sorter)

        for bangumi in data:
            bangumi["cover"] = "{}/{}".format(COVER_URL, normalize_path(bangumi["cover"]))

        data.reverse()

        for item in data:
            item["player"] = get_player(item["bangumi_name"])

        self.write(self.jsonify(data))
        self.finish()



from typing import Dict, Type

from bgmi.config import cfg
from bgmi.utils import print_error
from bgmi.website import bangumi_moe, base, mikan, share_dmhy

DATA_SOURCE_MAP: Dict[str, Type[base.BaseWebsite]] = {
    "mikan_project": mikan.Mikanani,
    "bangumi_moe": bangumi_moe.BangumiMoe,
    "dmhy": share_dmhy.DmhySource,
}

try:
    website = DATA_SOURCE_MAP[cfg.data_source]()
except KeyError:
    print_error(f'date source "{cfg.data_source}" in config is wrong, please edit it manually')



import os.path
import time
from operator import itemgetter
from typing import Any, Dict, List, Optional, Union

import filetype
import requests.exceptions
import sqlalchemy as sa

from bgmi.config import Source, cfg
from bgmi.lib.constants import BANGUMI_UPDATE_TIME, SUPPORT_WEBSITE
from bgmi.lib.download import Episode, download_prepare
from bgmi.lib.fetch import website
from bgmi.lib.table import (
    FOLLOWED_STATUS,
    STATUS_DELETED,
    STATUS_FOLLOWED,
    STATUS_NOT_DOWNLOAD,
    STATUS_UPDATED,
    Bangumi,
    Download,
    Filter,
    Followed,
    NotFoundError,
    Session,
    Subtitle,
    recreate_source_relatively_table,
)
from bgmi.script import ScriptRunner
from bgmi.utils import (
    COLOR_END,
    GREEN,
    convert_cover_url_to_path,
    download_cover,
    episode_filter_regex,
    logger,
    normalize_path,
    print_error,
    print_info,
    print_success,
    print_warning,
)

ControllerResult = Dict[str, Any]


def add(name: str, episode: Optional[int] = None) -> ControllerResult:
    """
    ret.name :str
    """
    # action add
    # add bangumi by a list of bangumi name
    logger.debug("add name: %s episode: %d", name, episode)
    if not Bangumi.get_updating_bangumi():
        website.fetch(group_by_weekday=False)

    try:
        bangumi_obj = Bangumi.fuzzy_get(name=name)
        name = bangumi_obj.name
    except Bangumi.NotFoundError:
        result = {
            "status": "error",
            "message": f"{name} not found, please check the name",
        }
        return result

    with Session.begin() as session:
        followed_obj: Optional[Followed] = session.scalar(
            sa.select(Followed).where(Followed.bangumi_name == bangumi_obj.name).limit(1)
        )
        if followed_obj is None:
            followed_obj = Followed(status=STATUS_FOLLOWED, bangumi_name=bangumi_obj.name, episode=0)
            session.add(followed_obj)
        else:
            if followed_obj.status == STATUS_FOLLOWED:
                result = {
                    "status": "warning",
                    "message": f"{bangumi_obj.name} already followed",
                }
                return result
            else:
                followed_obj.status = STATUS_FOLLOWED

        try:
            Filter.get(Filter.bangumi_name == name)
        except Filter.NotFoundError:
            session.add(Filter(bangumi_name=name))

    if episode is None:
        max_episode, _ = website.get_maximum_episode(bangumi_obj, max_page=cfg.max_path)
        followed_obj.episode = max_episode
    else:
        followed_obj.episode = episode

    with Session.begin() as session:
        session.add(followed_obj)

    result = {
        "status": "success",
        "message": f"{bangumi_obj.name} has been followed",
    }
    logger.debug(result)
    return result


def filter_(
    name: str,
    subtitle: Optional[str] = None,
    include: Optional[str] = None,
    exclude: Optional[str] = None,
    regex: Optional[str] = None,
) -> ControllerResult:
    result = {"status": "success", "message": ""}  # type: Dict[str, Any]
    try:
        bangumi_obj = Bangumi.fuzzy_get(name=name)
    except Bangumi.NotFoundError:
        result["status"] = "error"
        result["message"] = f"Bangumi {name} does not exist."
        return result

    if not Followed.get(Followed.bangumi_name == bangumi_obj.name):
        result["status"] = "error"
        result["message"] = "Bangumi {name} has not subscribed, try 'bgmi add \"{name}\"'.".format(
            name=bangumi_obj.name
        )
        return result

    followed_filter_obj = Filter.get(Filter.bangumi_name == bangumi_obj.name)

    if subtitle is not None:
        _subtitle = [s.strip() for s in subtitle.split(",")]
        _subtitle = [s.id for s in Subtitle.get_subtitle_by_name(_subtitle)]
        subtitle_list = [s.split(".")[0] for s in bangumi_obj.subtitle_group.split(", ") if "." in s]
        subtitle_list.extend(bangumi_obj.subtitle_group.split(", "))
        followed_filter_obj.subtitle = ", ".join(filter(lambda s: s in subtitle_list, _subtitle))

    if include is not None:
        followed_filter_obj.include = include

    if exclude is not None:
        followed_filter_obj.exclude = exclude

    if regex is not None:
        followed_filter_obj.regex = regex

    followed_filter_obj.save()
    subtitle_list = [s.name for s in Subtitle.get_subtitle_by_id(bangumi_obj.subtitle_group.split(", "))]

    result["data"] = {
        "name": bangumi_obj.name,
        "subtitle_group": subtitle_list,
        "followed": [s.name for s in Subtitle.get_subtitle_by_id(followed_filter_obj.subtitle.split(", "))]
        if followed_filter_obj.subtitle
        else [],
        "include": followed_filter_obj.include,
        "exclude": followed_filter_obj.exclude,
        "regex": followed_filter_obj.regex,
    }
    logger.debug(result)
    return result


def delete(name: str = "", clear_all: bool = False, batch: bool = False) -> ControllerResult:
    """
    :param name:
    :param clear_all:
    :param batch:
    """
    # action delete
    # just delete subscribed bangumi or clear all the subscribed bangumi
    result = {}
    logger.debug("delete %s", name)
    if clear_all:
        if Followed.delete_followed(batch=batch):
            result["status"] = "warning"
            result["message"] = "all subscriptions have been deleted"
        else:
            print_error("user canceled")
    elif name:
        try:
            followed = Followed.get(Followed.bangumi_name == name)
            followed.status = STATUS_DELETED
            followed.save()
            result["status"] = "warning"
            result["message"] = f"Bangumi {name} has been deleted"
        except Followed.NotFoundError:
            result["status"] = "error"
            result["message"] = f"Bangumi {name} does not exist"
    else:
        result["status"] = "warning"
        result["message"] = "Nothing has been done."

    logger.debug(result)

    return result


def cal(force_update: bool = False, cover: Optional[List[str]] = None) -> Dict[str, List[Dict[str, Any]]]:
    logger.debug("cal force_update: {}", force_update)

    weekly_list = Bangumi.get_updating_bangumi()
    if not weekly_list:
        print_warning("Warning: no bangumi schedule, fetching ...")
        force_update = True

    if force_update:
        print_info("Fetching bangumi info ...")
        website.fetch()

    weekly_list = Bangumi.get_updating_bangumi()

    if cover is not None:
        # download cover to local
        cover_to_be_download = cover
        for daily_bangumi in weekly_list.values():
            for bangumi in daily_bangumi:
                _, file_path = convert_cover_url_to_path(bangumi["cover"])

                if not (os.path.exists(file_path) and filetype.is_image(file_path)):
                    cover_to_be_download.append(bangumi["cover"])

        if cover_to_be_download:
            print_info("Updating cover ...")
            download_cover(cover_to_be_download)

    runner = ScriptRunner()
    patch_list = runner.get_models_dict()
    for i in patch_list:
        weekly_list[i["update_time"].lower()].append(i)
    logger.debug(weekly_list)

    # for web api, return all subtitle group info
    r = weekly_list  # type: Dict[str, List[Dict[str, Any]]]
    for day, value in weekly_list.items():
        for index, bangumi in enumerate(value):
            bangumi["cover"] = normalize_path(bangumi["cover"])
            subtitle_group = [
                {"name": x.name, "id": x.id}
                for x in Subtitle.get_subtitle_by_id(bangumi["subtitle_group"].split(", " ""))
            ]

            r[day][index]["subtitle_group"] = subtitle_group
    logger.debug(r)
    return r


def download(name: str, title: str, episode: int, download_url: str) -> None:
    download_prepare([Episode(name=name, title=title, episode=episode, download=download_url)])


def mark(name: str, episode: int) -> ControllerResult:
    """

    :param name: name of the bangumi you want to mark
    :param episode: bangumi episode you want to mark
    """
    result = {}

    try:
        followed_obj = Followed.get(Followed.bangumi_name == name)
    except Followed.NotFoundError:
        runner = ScriptRunner()
        followed_obj = runner.get_model(name)  # type: ignore
        if not followed_obj:
            result["status"] = "error"
            result["message"] = f"Subscribe or Script <{name}> does not exist."
            return result

    followed_obj.episode = episode
    followed_obj.save()

    result["status"] = "success"
    result["message"] = f"{name} has been mark as episode: {episode}"
    return result


def search(
    keyword: str,
    count: Union[str, int] = cfg.max_path,
    regex: Optional[str] = None,
    dupe: bool = False,
    min_episode: Optional[int] = None,
    max_episode: Optional[int] = None,
    tag: bool = False,
    subtitle: Optional[str] = None,
) -> ControllerResult:
    try:
        count = int(count)
    except (TypeError, ValueError):
        count = 3
    try:
        if tag:
            data = website.search_by_tag(keyword, subtitle=subtitle, count=count)
        else:
            data = website.search_by_keyword(keyword, count=count)
        data = episode_filter_regex(data, regex=regex)
        if min_episode is not None:
            data = [x for x in data if x.episode >= min_episode]
        if max_episode is not None:
            data = [x for x in data if x.episode <= max_episode]

        if not dupe:
            data = Episode.remove_duplicated_bangumi(data)
        data.sort(key=lambda x: x.episode)
        return {
            "status": "success",
            "message": "",
            "options": {
                "keyword": keyword,
                "count": count,
                "regex": regex,
                "dupe": dupe,
                "min_episode": min_episode,
                "max_episode": max_episode,
            },
            "data": data,
        }
    except Exception as e:
        if os.environ.get("DEBUG"):
            raise
        return {
            "status": "error",
            "message": str(e),
            "options": {
                "keyword": keyword,
                "count": count,
                "regex": regex,
                "dupe": dupe,
                "min_episode": min_episode,
                "max_episode": max_episode,
            },
            "data": [],
        }


def source(data_source: str) -> ControllerResult:
    result = {}
    if data_source in list(map(itemgetter("id"), SUPPORT_WEBSITE)):
        recreate_source_relatively_table()
        cfg.data_source = Source(data_source)
        cfg.save()
        print_success("data source switch succeeds")
        result["status"] = "success"
        result["message"] = f"you have successfully change your data source to {data_source}"
    else:
        result["status"] = "error"
        result["message"] = "please check your input. data source should be one of {}".format(
            [x["id"] for x in SUPPORT_WEBSITE]
        )
    return result


def update(names: List[str], download: Optional[bool] = False, not_ignore: bool = False) -> ControllerResult:
    logger.debug("updating bangumi info with args: download: %r", download)
    downloaded: List[Episode] = []
    result: Dict[str, Any] = {
        "status": "info",
        "message": "",
        "data": {"updated": [], "downloaded": downloaded},
    }

    ignore = not bool(not_ignore)
    print_info("marking bangumi status ...")
    now = int(time.time())

    for follow in Followed.get_all_followed():
        if follow.updated_time and int(follow.updated_time + 60 * 60 * 24) < now:
            followed_obj = Followed.get(Followed.bangumi_name == follow.bangumi_name)
            followed_obj.status = STATUS_FOLLOWED
            followed_obj.save()

    for script in ScriptRunner().scripts:
        obj = script.Model().obj
        if obj.updated_time and int(obj.updated_time + 60 * 60 * 24) < now:
            obj.status = STATUS_FOLLOWED
            obj.save()

    print_info("updating subscriptions ...")

    if not names:
        updated_bangumi_obj = Followed.get_all_followed()
    else:
        updated_bangumi_obj = []
        for n in names:
            try:
                f = Followed.get(Followed.bangumi_name == n)
                updated_bangumi_obj.append(f)
            except Followed.NotFoundError:
                logger.warning("missing followed bangumi '{}'", n)

    runner = ScriptRunner()
    script_download_queue = runner.run()
    if script_download_queue and download:
        download_prepare(script_download_queue)
        downloaded.extend(script_download_queue)
        print_info("downloading ...")
        download_prepare(
            [
                Episode(**{key: value for key, value in x.items() if key not in ["id", "status"]})
                for x in Download.get_all_downloads(status=STATUS_NOT_DOWNLOAD)
            ]
        )

    for subscribe in updated_bangumi_obj:
        download_queue = []
        print_info(f"fetching {subscribe.bangumi_name} ...")
        try:
            bangumi_obj = Bangumi.get(Bangumi.name == subscribe.bangumi_name)
        except NotFoundError:
            logger.error("Bangumi<{}> does not exists.", subscribe.bangumi_name)
            continue
        try:
            followed_obj = Followed.get(Followed.bangumi_name == subscribe.bangumi_name)
        except NotFoundError:
            logger.error("Followed<{}> is not followed.", subscribe.bangumi_name)
            continue

        try:
            episode, all_episode_data = website.get_maximum_episode(
                bangumi=bangumi_obj, ignore_old_row=ignore, max_page=cfg.max_path
            )
        except requests.exceptions.ConnectionError as e:
            print_warning(f"error {e} to fetch {bangumi_obj.name}, skip")
            continue

        # TODO(v5): add default to column and remove this
        saved_episode = subscribe.episode or 0
        if episode > saved_episode:
            episode_range = range(saved_episode, episode + 1)
            print_success(f"{subscribe.bangumi_name} updated, episode: {episode:d}")
            followed_obj.episode = episode
            followed_obj.status = STATUS_UPDATED
            followed_obj.updated_time = int(time.time())
            followed_obj.save()
            result["data"]["updated"].append({"bangumi": subscribe.bangumi_name, "episode": episode})

            for i in episode_range:
                for epi in all_episode_data:
                    if epi.episode == i:
                        download_queue.append(epi)
                        break

        if download:
            download_prepare(download_queue)
            downloaded.extend(download_queue)

    if downloaded:
        failed = [Episode.parse_obj(x) for x in Download.get_all_downloads(status=STATUS_NOT_DOWNLOAD)]
        if failed:
            print_info("try to re-downloading previous failed torrents ...")
            download_prepare(failed)

    return result


def status_(name: str, status: int = STATUS_DELETED) -> ControllerResult:
    result = {"status": "success", "message": ""}

    if (status not in FOLLOWED_STATUS) or (not status):
        result["status"] = "error"
        result["message"] = f"Invalid status: {status}"
        return result

    status = int(status)

    followed_obj = Followed.get(Followed.bangumi_name == name)
    if not followed_obj:
        result["status"] = "error"
        result["message"] = f"Followed<{name}> does not exists"
        return result

    followed_obj.status = status
    followed_obj.save()
    result["message"] = f"Followed<{name}> has been marked as status {status}"
    return result


def list_() -> ControllerResult:
    result = {}
    weekday_order = BANGUMI_UPDATE_TIME
    followed_bangumi = website.followed_bangumi()

    script_bangumi = ScriptRunner().get_models_dict()

    if not followed_bangumi and not script_bangumi:
        result["status"] = "warning"
        result["message"] = "you have not subscribed any bangumi"
        return result

    for i in script_bangumi:
        i["subtitle_group"] = [{"name": "<BGmi Script>"}]
        followed_bangumi[i["update_time"].lower()].append(i)

    result["status"] = "info"
    result["message"] = ""
    for weekday in weekday_order:
        if followed_bangumi[weekday.lower()]:
            result["message"] += f"{GREEN}{weekday}. {COLOR_END}"
            for j, bangumi in enumerate(followed_bangumi[weekday.lower()]):
                if bangumi["status"] in (STATUS_UPDATED, STATUS_FOLLOWED) and "episode" in bangumi:
                    bangumi["name"] = f"{bangumi['name']}({bangumi['episode']:d})"
                if j > 0:
                    result["message"] += " " * 5
                f = [x["name"] for x in bangumi["subtitle_group"]]
                result["message"] += "{}: {}\n".format(bangumi["name"], ", ".join(f) if f else "<None>")

    return result



import functools
import glob
import gzip
import json
import os
import re
import struct
import subprocess
import sys
import tarfile
import time
import traceback
from io import BytesIO
from multiprocessing.pool import ThreadPool
from pathlib import Path
from shutil import move, rmtree
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar

import requests
import semver
from anime_episode_parser import parse_episode as _parse_episode
from loguru import logger
from requests import Response

from bgmi import __admin_version__, __version__
from bgmi.config import BGMI_PATH, IS_WINDOWS, cfg
from bgmi.lib.constants import SUPPORT_WEBSITE
from bgmi.session import session
from bgmi.website.model import Episode

F = TypeVar("F", bound=Callable[..., Any])

if (
    IS_WINDOWS and "bash" not in os.getenv("SHELL", "").lower() and "zsh" not in os.getenv("SHELL", "").lower()
):  # pragma: no cover
    GREEN = ""
    YELLOW = ""
    RED = ""
    COLOR_END = ""
else:
    GREEN = "\033[1;32m"
    YELLOW = "\033[1;33m"
    RED = "\033[1;31m"
    COLOR_END = "\033[0m"

color_map = {
    "print_info": "",
    "print_success": GREEN,
    "print_warning": YELLOW,
    "print_error": RED,
}

indicator_map = {
    "print_info": "[*] ",
    "print_success": "[+] ",
    "print_warning": "[-] ",
    "print_error": "[x] ",
}


def print_info(message: str, indicator: bool = True) -> None:
    logger.info(message)


def print_success(message: str, indicator: bool = True) -> None:
    logger.info(message)


def print_warning(message: str, indicator: bool = True) -> None:
    logger.warning(message)


def print_error(message: str, stop: bool = True, indicator: bool = True) -> None:
    logger.error(message)
    if stop:
        sys.exit(1)


def print_version() -> str:
    return """BGmi {c}ver. {v}{e} built by {c}RicterZ{e} with ❤️

Github: https://github.com/BGmi/BGmi
Email: ricterzheng@gmail.com
Blog: https://ricterz.me""".format(
        c=YELLOW, v=__version__, e=COLOR_END
    )


def test_connection() -> bool:
    try:
        for website in SUPPORT_WEBSITE:
            if cfg.data_source == website["id"]:
                session.request("head", website["url"], timeout=10)
    except requests.RequestException:
        return False
    return True


def bug_report() -> None:  # pragma: no cover
    print_error(
        "It seems that no bangumi found, if https://bangumi.moe can \n    be opened "
        "normally, please submit issue at: https://github.com/BGmi/BGmi/issues"
    )


_DEFAULT_TERMINAL_WIDTH = 80


def get_terminal_col() -> int:  # pragma: no cover
    # pylint: disable=import-outside-toplevel,import-error
    # https://gist.github.com/jtriley/1108174
    if not IS_WINDOWS:
        import fcntl
        import termios

        try:
            col = struct.unpack(
                "HHHH",
                fcntl.ioctl(0, termios.TIOCGWINSZ, struct.pack("HHHH", 0, 0, 0, 0)),
            )[
                1
            ]  # type: int

            return col
        except Exception:
            return _DEFAULT_TERMINAL_WIDTH
    else:
        try:
            from ctypes import create_string_buffer, windll  # type: ignore[attr-defined]

            # stdin handle is -10
            # stdout handle is -11
            # stderr handle is -12
            h = windll.kernel32.GetStdHandle(-12)
            csbi = create_string_buffer(22)
            res = windll.kernel32.GetConsoleScreenBufferInfo(h, csbi)
            if res:
                (
                    bufx,
                    bufy,
                    curx,
                    cury,
                    wattr,
                    left,
                    top,
                    right,
                    bottom,
                    maxx,
                    maxy,
                ) = struct.unpack("hhhhHhhhhhh", csbi.raw)
                sizex = right - left + 1  # type: int
                return sizex
            else:
                cols = int(subprocess.check_output("tput cols"))
                return cols
        except Exception:
            return _DEFAULT_TERMINAL_WIDTH


FRONTEND_NPM_URL = "https://registry.npmjs.com/bgmi-frontend/"


@functools.lru_cache
def npm_package_manifest() -> Dict[str, Any]:
    r = session.get(FRONTEND_NPM_URL, timeout=60)
    r.raise_for_status()
    return r.json()  # type: ignore


@functools.lru_cache
def latest_npm_package_version() -> semver.VersionInfo:
    r = npm_package_manifest()
    versions = sorted([semver.VersionInfo.parse(x) for x in r["versions"]])

    available_versions = [v for v in versions if all(v.match(r) for r in __admin_version__)]

    if not available_versions:
        print_error("failed to find available web-ui versions")

    return max(available_versions)


def check_update(mark: bool = True) -> None:
    def update() -> None:
        try:
            print_info("Checking update ...")
            pypi = session.get("https://pypi.org/pypi/bgmi/json", timeout=60).json()
            version = pypi["info"]["version"]

            with open(os.path.join(BGMI_PATH, "latest"), "w", encoding="utf8") as f:
                f.write(version)

            if version > __version__:
                print_warning(
                    "Please update bgmi to the latest version {}{}{}."
                    "\nThen execute `bgmi upgrade` to migrate database".format(GREEN, version, COLOR_END)
                )
            else:
                print_success("Your BGmi is the latest version.")

            if cfg.front_static_path.joinpath("package.json").exists():
                admin_version = latest_npm_package_version()

                with open(os.path.join(cfg.front_static_path, "package.json"), encoding="utf8") as f:
                    local_version = semver.VersionInfo.parse(json.loads(f.read())["version"])

                if admin_version > local_version:
                    get_web_admin(method="update")
            else:
                print_info("Use 'bgmi install' to install BGmi frontend / download delegate")
            if not mark:
                update()
                raise SystemExit
        except Exception as e:
            print_warning(f"Error occurs when checking update, {str(e)}")

    version_file = os.path.join(BGMI_PATH, "version")
    if not os.path.exists(version_file):
        with open(version_file, "w", encoding="utf8") as f:
            f.write(str(int(time.time())))
        update()

    with open(version_file, encoding="utf8") as f:
        try:
            data = int(f.read())
            if time.time() - 7 * 24 * 3600 > data:
                with open(version_file, "w", encoding="utf8") as f:
                    f.write(str(int(time.time())))
                update()
        except ValueError:
            pass


def parse_episode(episode_title: str) -> int:
    s, c = _parse_episode(episode_title)
    if c != 1:
        return 0

    return s or 0


def normalize_path(url: str) -> str:
    """
    normalize link to path

    :param url: path or url to normalize
    :type url: str
    :return: normalized path
    :rtype: str
    """
    url = url.replace("http://", "http/").replace("https://", "https/")
    illegal_char = [":", "*", "?", '"', "<", ">", "|", "'"]
    for char in illegal_char:
        url = url.replace(char, "")

    if url.startswith("/"):
        return url[1:]
    else:
        return url


def bangumi_save_path(bangumi_name: str) -> Path:
    name = normalize_path(bangumi_name)

    r = cfg.save_path_map.get(name)
    if r is None:
        return cfg.save_path.joinpath(name)

    if r.is_absolute():
        return r

    return cfg.save_path.joinpath(r)


def get_web_admin(method: str) -> None:
    print_info(f"{method[0].upper() + method[1:]}ing BGmi frontend")
    admin_version = latest_npm_package_version()

    try:
        r = npm_package_manifest()
        version = r["versions"][str(admin_version)]
        tar_url = version["dist"]["tarball"]
    except requests.exceptions.ConnectionError:
        print_warning("failed to download web admin")
        return
    except json.JSONDecodeError:
        print_warning("failed to download web admin")
        return

    tar = session.get(tar_url, timeout=60)
    tar.raise_for_status()
    admin_zip = BytesIO(tar.content)
    with gzip.GzipFile(fileobj=admin_zip) as f:
        tar_file = BytesIO(f.read())

    rmtree(cfg.front_static_path)
    os.makedirs(cfg.front_static_path)

    with tarfile.open(fileobj=tar_file) as tar_file_obj:
        tar_file_obj.extractall(path=cfg.front_static_path)

    for file in os.listdir(os.path.join(cfg.front_static_path, "package", "dist")):
        move(
            os.path.join(cfg.front_static_path, "package", "dist", file),
            os.path.join(cfg.front_static_path, file),
        )

    with open(os.path.join(cfg.front_static_path, "package.json"), "w+", encoding="utf8") as pkg:
        pkg.write(json.dumps(version, ensure_ascii=False, indent=2))
    print_success("Web admin page {} successfully. version: {}".format(method, version["version"]))


def convert_cover_url_to_path(cover_url: str) -> Tuple[str, str]:
    """
    convert bangumi cover to file path

    :param cover_url: bangumi cover path
    :type cover_url:str
    :rtype: str,str
    :return: dir_path, file_path
    """

    cover_url = normalize_path(cover_url)
    file_path = os.path.join(cfg.save_path, "cover")
    file_path = os.path.join(file_path, cover_url)
    dir_path = os.path.dirname(file_path)

    return dir_path, file_path


def download_file(url: str) -> Optional[Response]:
    logger.debug("downloading {}", url)
    if url.startswith("https://") or url.startswith("http://"):
        print_info(f"Download: {url}")
        return session.get(url, timeout=60)
    return None


def download_cover(cover_url_list: List[str]) -> None:
    p = ThreadPool(4)
    content_list = p.map(download_file, cover_url_list)
    p.close()
    for index, r in enumerate(content_list):
        if not r:
            continue

        dir_path, file_path = convert_cover_url_to_path(cover_url_list[index])
        if not glob.glob(dir_path):
            os.makedirs(dir_path)
        with open(file_path, "wb") as f:
            f.write(r.content)


def episode_filter_regex(data: List[Episode], regex: Optional[str] = None) -> List[Episode]:
    """

    :param data: list of bangumi dict
    :param regex: regex
    """
    if regex:
        try:
            match = re.compile(regex)
            data = [s for s in data if match.findall(s.title)]
        except re.error as e:
            if os.getenv("DEBUG"):  # pragma: no cover
                traceback.print_exc()
                raise e
            print_warning(f"can't compile regex {regex}, skipping filter by regex")

    if cfg.enable_global_filters:
        exclude_keywords = [t.strip().lower() for t in cfg.global_filters]
        data = [x for x in data if not x.contains_any_words(exclude_keywords)]

    return data



from .aria2_rpc import Aria2DownloadRPC
from .deluge import DelugeRPC
from .qbittorrent import QBittorrentWebAPI
from .transmission import TransmissionRPC

__all__ = ["Aria2DownloadRPC", "DelugeRPC", "QBittorrentWebAPI", "TransmissionRPC"]



import sqlite3
from pathlib import Path

import packaging.version

from bgmi import __version__
from bgmi.config import BGMI_PATH, cfg
from bgmi.utils import COLOR_END, RED, print_error, print_info

old_version_file = BGMI_PATH.joinpath("old")


def exec_sql(sql: str, db: Path = cfg.db_path) -> None:
    try:
        print_info(f"Execute {sql}")
        conn = sqlite3.connect(db)
        conn.execute(sql)
        conn.commit()
        conn.close()
    except sqlite3.OperationalError:  # pragma: no cover
        print_error("Execute SQL statement failed", stop=False)


def update_database() -> None:
    if not old_version_file.exists():
        old_version_file.write_text(__version__, encoding="utf8")
        return

    previous = packaging.version.parse(old_version_file.read_text(encoding="utf8"))
    if previous < packaging.version.Version("4.0.0"):
        print_error(
            (
                "can't automatically upgrade from <4.0.0 version, "
                + " please backup your .bgmi files, remove them and use `bgmi install`\n"
                + RED
                + "All Data will lost, you will need to re-add your bangumi after re-install"
                + COLOR_END
            ),
            stop=True,
        )

    # all upgrade done, write current version
    old_version_file.write_text(__version__, encoding="utf8")



import os
import time
from collections import defaultdict
from typing import Any, Dict, List, Optional, Type, TypeVar

import sqlalchemy as sa
from sqlalchemy import CHAR, Column, Integer, Text, create_engine
from sqlalchemy.orm import DeclarativeBase, Mapped, sessionmaker

from bgmi.config import cfg
from bgmi.utils import episode_filter_regex
from bgmi.website.model import Episode

# bangumi status
STATUS_UPDATING = 0
STATUS_END = 1
BANGUMI_STATUS = (STATUS_UPDATING, STATUS_END)

# subscription status
STATUS_DELETED = 0
STATUS_FOLLOWED = 1
STATUS_UPDATED = 2
FOLLOWED_STATUS = (STATUS_DELETED, STATUS_FOLLOWED, STATUS_UPDATED)

# download status
STATUS_NOT_DOWNLOAD = 0
STATUS_DOWNLOADING = 1
STATUS_DOWNLOADED = 2
DOWNLOAD_STATUS = (STATUS_NOT_DOWNLOAD, STATUS_DOWNLOADING, STATUS_DOWNLOADED)

engine = create_engine(f"sqlite:///{cfg.db_path.absolute().as_posix()}")
# if os.getenv("DEV")
# engine = create_engine("sqlite:///./tmp/bangumi.db")

Session = sessionmaker(engine, expire_on_commit=False)


class NotFoundError(Exception):
    def __init__(self, cls: Type["Base"]) -> None:
        super().__init__(f"{cls} not found")


T = TypeVar("T", bound="Base")


class Base(DeclarativeBase):
    NotFoundError = NotFoundError

    @classmethod
    def get(cls: Type[T], *where: Any) -> T:
        with Session.begin() as session:
            o = session.scalar(sa.select(cls).where(*where).limit(1))
            if not o:
                raise NotFoundError(cls)
            return o

    @classmethod
    def all(cls: Type[T], *where: Any) -> List[T]:
        with Session.begin() as session:
            return list(session.scalars(sa.select(cls).where(*where)).all())

    def save(self) -> None:
        with Session.begin() as session:
            session.add(self)


metadata = Base.metadata
if os.environ.get("DEV"):
    print(f"using database {cfg.db_path}")


class Bangumi(Base):
    __tablename__ = "bangumi"
    STATUS_UPDATING = 0
    STATUS_END = 1

    id: Mapped[int] = Column(Integer, primary_key=True)  # type: ignore
    name: Mapped[str] = Column(Text, nullable=False, unique=True)  # type: ignore
    subtitle_group: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    keyword: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    update_time: Mapped[str] = Column(
        CHAR(5), nullable=False, default="Unknown", server_default="Unknown"
    )  # type: ignore
    cover: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    status: Mapped[int] = Column(
        Integer, nullable=False, default=STATUS_UPDATING, server_default=str(STATUS_UPDATING)
    )  # type: ignore

    @classmethod
    def delete_all(cls) -> None:
        with Session.begin() as session:
            un_updated_bangumi: List[Followed] = list(
                session.scalars(
                    sa.select(Followed).where(Followed.updated_time > (int(time.time()) - 2 * 7 * 24 * 3600))
                )
            )

            if os.getenv("DEBUG"):  # pragma: no cover
                print("ignore updating bangumi", [x.bangumi_name for x in un_updated_bangumi])

            # do not mark updating bangumi as STATUS_END
            session.execute(
                sa.update(cls)
                .where(cls.name.not_in([x.bangumi_name for x in un_updated_bangumi]))
                .values(status=STATUS_END)
            )

    @classmethod
    def get_updating_bangumi(
        cls,
        status: Optional[int] = None,
    ) -> Dict[str, List[Dict[str, Any]]]:
        if status is None:
            sql = (
                sa.select(cls, Followed.status, Followed.episode)
                .outerjoin(Followed, cls.name == Followed.bangumi_name)
                .where(cls.status == STATUS_UPDATING)
            )
        else:
            sql = (
                sa.select(cls, Followed.status, Followed.episode)
                .outerjoin(Followed, cls.name == Followed.bangumi_name)
                .where((cls.status == STATUS_UPDATING) & (Followed.status == status))
            )

        with Session() as session:
            # if status is None:
            data = session.scalars(sql).all()

        weekly_list = defaultdict(list)
        for bangumi_item in data:
            weekly_list[bangumi_item.update_time.lower()].append(bangumi_item.__dict__)

        return weekly_list

    @classmethod
    def fuzzy_get(cls, name: str) -> "Bangumi":
        return cls.get(cls.name.contains(name))


class Followed(Base):
    __tablename__ = "followed"

    STATUS_DELETED = 0
    STATUS_FOLLOWED = 1
    STATUS_UPDATED = 2

    id: Mapped[int] = Column(Integer, primary_key=True)  # type: ignore
    bangumi_name: Mapped[str] = Column(Text, nullable=False, unique=True)  # type: ignore
    episode: Mapped[int] = Column(Integer, default=0, server_default="0")  # type: ignore
    status: Mapped[int] = Column(Integer)  # type: ignore
    updated_time: Mapped[int] = Column(Integer, default=0, server_default="0")  # type: ignore

    @classmethod
    def delete_followed(cls, batch: bool = True) -> bool:
        if not batch:
            if not input("[+] are you sure want to CLEAR ALL THE BANGUMI? (y/N): ") == "y":
                return False

        with Session.begin() as session:
            session.execute(sa.delete(cls))
        return True

    @classmethod
    def get_all_followed(
        cls: Type["Followed"], status: int = STATUS_DELETED, bangumi_status: int = STATUS_UPDATING
    ) -> List["Followed"]:
        sql = (
            sa.select(Followed)
            .join(Bangumi, cls.bangumi_name == Bangumi.name)
            .where(cls.status.isnot(status), Bangumi.status == bangumi_status)
            .order_by(cls.updated_time.desc())
        )

        with Session() as s:
            return list(s.scalars(sql))


class Download(Base):
    __tablename__ = "download"

    id: Mapped[int] = Column(Integer, primary_key=True)  # type: ignore
    name: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    title: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    episode: Mapped[int] = Column(Integer, nullable=False)  # type: ignore
    download: Mapped[str] = Column(Text, nullable=False)  # type: ignore
    status: Mapped[int] = Column(Integer, nullable=False)  # type: ignore

    @classmethod
    def get_all_downloads(cls, status: Optional[int] = None) -> List[dict]:
        with Session.begin() as session:
            if status is None:
                sql = sa.select(cls).where().order_by(cls.status)
            else:
                sql = sa.select(cls).where(cls.status == status).order_by(cls.status)

            return [x.__dict__ for x in session.scalars(sql)]

    def downloaded(self) -> None:
        self.status = STATUS_DOWNLOADED
        self.save()


class Filter(Base):
    __tablename__ = "filter"

    id: Mapped[int] = Column(Integer, primary_key=True)  # type: ignore
    bangumi_name: Mapped[str] = Column(Text, nullable=False, unique=True)  # type: ignore
    subtitle: Mapped[str] = Column(Text)  # type: ignore
    include: Mapped[str] = Column(Text)  # type: ignore
    exclude: Mapped[str] = Column(Text)  # type: ignore
    regex: Mapped[str] = Column(Text)  # type: ignore

    @property
    def subtitle_group_split(self) -> List[str]:
        if self.subtitle:
            # pylint:disable=no-member
            return [x.strip() for x in self.subtitle.split(",")]
        else:
            return []

    def apply_on_episodes(self, result: List[Episode]) -> List[Episode]:
        if self.include:
            # pylint:disable=no-member
            include_list = [s.strip().lower() for s in self.include.split(",")]
            result = [e for e in result if e.contains_any_words(include_list)]

        if cfg.enable_global_include_keywords:
            include_list = [s.strip().lower() for s in cfg.global_include_keywords]
            result = [e for e in result if e.contains_any_words(include_list)]

        if self.exclude:
            # pylint:disable=no-member
            exclude_list = [s.strip().lower() for s in self.exclude.split(",")]
            result = [e for e in result if not e.contains_any_words(exclude_list)]

        return episode_filter_regex(data=result, regex=self.regex)


class Subtitle(Base):
    __tablename__ = "subtitle"

    id: Mapped[str] = Column(Text, primary_key=True)  # type: ignore
    name: Mapped[str] = Column(Text, nullable=False)  # type: ignore

    @classmethod
    def get_subtitle_by_id(cls, id_list: List[str]) -> List["Subtitle"]:
        with Session.begin() as session:
            return list(session.scalars(sa.select(cls).where(cls.id.in_(id_list))))

    @classmethod
    def get_subtitle_by_name(cls, name_list: List[str]) -> List["Subtitle"]:
        with Session.begin() as session:
            return list(session.scalars(sa.select(cls).where(cls.name.in_(name_list))))


class Scripts(Base):
    __tablename__ = "scripts"

    id: Mapped[int] = Column(Integer, primary_key=True)  # type: ignore
    bangumi_name: Mapped[str] = Column(Text, nullable=False, unique=True)  # type: ignore
    episode: Mapped[int] = Column(Integer, nullable=False)  # type: ignore
    status: Mapped[int] = Column(Integer, nullable=False)  # type: ignore
    updated_time: Mapped[int] = Column(Integer, nullable=False, default=0, server_default="0")  # type: ignore


def recreate_source_relatively_table() -> None:
    with Session.begin() as session:
        for table in [Subtitle, Filter, Download, Followed, Bangumi]:
            session.execute(sa.delete(table))


def recreate_scripts_table() -> None:
    with Session.begin() as session:
        for table in [
            Scripts,
        ]:
            session.execute(sa.delete(table))



from typing import List

SUPPORT_WEBSITE = [
    {"view": "萌番组 https://bangumi.moe/", "id": "bangumi_moe"},
    {"view": "蜜柑计划 https://mikanani.me/", "id": "mikan_project"},
    {
        "view": "动漫花园 http://share.dmhy.org/",
        "id": "dmhy",
    },
]

SPACIAL_APPEND_CHARS = ["Ⅱ", "Ⅲ", "♪", "Δ", "×", "☆", "É", "·", "♭", "★"]
SPACIAL_REMOVE_CHARS: List[str] = []

BANGUMI_UPDATE_TIME = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Unknown")
